@@ -6,7 +6,8 @@ from __future__ import unicode_literals
 
 import sys, codecs
 from datetime import datetime, timedelta, tzinfo
-
+from copy import copy
+import codecs
 from .exceptions import EncodingError
 
 # For Python 3 support
@@ -15,12 +16,14 @@ if PYTHON_VERSION >= 3:
     MAX_INT = sys.maxsize
     dictItemsIter = dict.items
     xrange = range
-    #toByteArray = lambda x: bytearray(codecs.decode(x, 'hex_codec')) if type(x) in (bytes, str) else x
+    unichr = chr
     toByteArray = lambda x: bytearray(codecs.decode(x, 'hex_codec')) if type(x) == bytes else bytearray(codecs.decode(bytes(x, 'ascii'), 'hex_codec')) if type(x)  == str else x
-else:
+    rawStrToByteArray = lambda x: bytearray(bytes(x, 'latin-1'))
+else: #pragma: no cover
     MAX_INT = sys.maxint
     dictItemsIter = dict.iteritems
     toByteArray = lambda x: bytearray(x.decode('hex')) if type(x) in (str, unicode) else x
+    rawStrToByteArray = bytearray
 
 # Tables can be found at: http://en.wikipedia.org/wiki/GSM_03.38#GSM_7_bit_default_alphabet_and_extension_table_of_3GPP_TS_23.038_.2F_GSM_03.38
 GSM7_BASIC = ('@£$¥èéùìòÇ\nØø\rÅåΔ_ΦΓΛΩΠΨΣΘΞ\x1bÆæßÉ !\"#¤%&\'()*+,-./0123456789:;<=>?¡ABCDEFGHIJKLMNOPQRSTUVWXYZÄÖÑÜ`¿abcdefghijklmnopqrstuvwxyzäöñüà')
@@ -36,6 +39,10 @@ GSM7_EXTENDED = {chr(0xFF): 0x0A,
                  ']':  chr(0x3E),
                  '|':  chr(0x40),
                  '€':  chr(0x65)}
+# Maximum message sizes for each data coding
+MAX_MESSAGE_LENGTH = {0x00: 160, # GSM-7
+                      0x04: 140, # 8-bit
+                      0x08: 70}  # UCS2
 
 class SmsPduTzInfo(tzinfo):
     """ Simple implementation of datetime.tzinfo for handling timestamp GMT offsets specified in SMS PDUs """
@@ -54,19 +61,176 @@ class SmsPduTzInfo(tzinfo):
             self._setPduOffsetStr(pduOffsetStr)
         
     def _setPduOffsetStr(self, pduOffsetStr):
-        # See if the timezone difference is positive/negative               
-        if int(pduOffsetStr, 16) & 0x80 == 0: # positive            
+        # See if the timezone difference is positive/negative by checking MSB of first semi-octet
+        tzHexVal = int(pduOffsetStr, 16)
+        if tzHexVal & 0x80 == 0: # positive
             self._offset = timedelta(minutes=(int(pduOffsetStr) * 15))
-        else: # negative            
-            self._offset = timedelta(minutes=(int(pduOffsetStr) * -15))
+        else: # negative
+            self._offset = timedelta(minutes=(int('{0:0>2X}'.format(tzHexVal & 0x7F)) * -15))
     
     def utcoffset(self, dt):
         return self._offset
     
     def dst(self, dt):
-        """ We do not have enough info in the SMS PDU to implement DST """
+        """ We do not have enough info in the SMS PDU to implement daylight savings time """
         return timedelta(0)
+
+
+class InformationElement(object):
+    """ User Data Header (UDH) Information Element (IE) implementation
+     
+    This represents a single field ("information element") in the PDU's
+    User Data Header. The UDH itself contains one or more of these
+    information elements.
     
+    If the IEI (IE identifier) is recognized, the class will automatically
+    specialize into one of the subclasses of InformationElement, 
+    e.g. Concatenation or PortAddress, allowing the user to easily
+    access the specific (and useful) attributes of these special cases.
+    """
+    
+    def __new__(cls, *args, **kwargs): #iei, ieLen, ieData):
+        """ Causes a new InformationElement class, or subclass
+        thereof, to be created. If the IEI is recognized, a specific
+        subclass of InformationElement is returned """
+        if len(args) > 0:
+            targetClass = IEI_CLASS_MAP.get(args[0], cls)
+        elif 'iei' in kwargs:
+            targetClass = IEI_CLASS_MAP.get(kwargs['iei'], cls)
+        else:
+            return super(InformationElement, cls).__new__(cls)
+        return super(InformationElement, targetClass).__new__(targetClass)
+    
+    def __init__(self, iei, ieLen=0, ieData=None):
+        self.id = iei # IEI
+        self.dataLength = ieLen # IE Length
+        self.data = ieData or [] # raw IE data
+        
+    @classmethod
+    def decode(cls, byteIter):
+        """ Decodes a single IE at the current position in the specified
+        byte iterator 
+        
+        @return: An InformationElement (or subclass) instance for the decoded IE
+        @rtype: InformationElement, or subclass thereof
+        """
+        iei = next(byteIter)
+        ieLen = next(byteIter)
+        ieData = []
+        for i in xrange(ieLen):
+            ieData.append(next(byteIter))
+        return InformationElement(iei, ieLen, ieData)
+    
+    def encode(self):
+        """ Encodes this IE and returns the resulting bytes """
+        result = bytearray()
+        result.append(self.id)
+        result.append(self.dataLength)
+        result.extend(self.data)
+        return result
+    
+    def __len__(self):
+        """ Exposes the IE's total length (including the IEI and IE length octet) in octets """
+        return self.dataLength + 2
+
+
+class Concatenation(InformationElement):
+    """ IE that indicates SMS concatenation.
+    
+    This implementation handles both 8-bit and 16-bit concatenation
+    indication, and exposes the specific useful details of this
+    IE as instance variables.
+    
+    Exposes:
+    reference: CSMS reference number, must be same for all the SMS parts in the CSMS
+    parts: total number of parts. The value shall remain constant for every short
+     message which makes up the concatenated short message. If the value is zero then 
+     the receiving entity shall ignore the whole information element
+    number:  this part's number in the sequence. The value shall start at 1 and
+     increment for every short message which makes up the concatenated short message
+    """
+    
+    def __init__(self, iei=0x00, ieLen=0, ieData=None):
+        super(Concatenation, self).__init__(iei, ieLen, ieData)
+        if ieData != None:
+            if iei == 0x00: # 8-bit reference
+                self.reference, self.parts, self.number = ieData
+            else: # 0x08: 16-bit reference
+                self.reference = ieData[0] << 8 | ieData[1]
+                self.parts = ieData[2]
+                self.number = ieData[3]
+
+    def encode(self):
+        if self.reference > 0xFF:
+            self.id = 0x08 # 16-bit reference
+            self.data = [self.reference >> 8, self.reference & 0xFF, self.parts, self.number]
+        else:
+            self.id = 0x00 # 8-bit reference
+            self.data = [self.reference, self.parts, self.number]
+        self.dataLength = len(self.data)
+        return super(Concatenation, self).encode()
+
+
+class PortAddress(InformationElement):
+    """ IE that indicates an Application Port Addressing Scheme.
+    
+    This implementation handles both 8-bit and 16-bit concatenation
+    indication, and exposes the specific useful details of this
+    IE as instance variables.
+    
+    Exposes:
+    destination: The destination port number
+    source: The source port number
+    """
+    
+    def __init__(self, iei=0x04, ieLen=0, ieData=None):
+        super(PortAddress, self).__init__(iei, ieLen, ieData)
+        if ieData != None:
+            if iei == 0x04: # 8-bit port addressing scheme
+                self.destination, self.source = ieData
+            else: # 0x05: 16-bit port addressing scheme
+                self.destination = ieData[0] << 8 | ieData[1]
+                self.source = ieData[2] << 8 | ieData[3]
+    
+    def encode(self):
+        if self.destination > 0xFF or self.source > 0xFF:
+            self.id = 0x05 # 16-bit
+            self.data = [self.destination >> 8, self.destination & 0xFF, self.source >> 8, self.source & 0xFF]
+        else:
+            self.id = 0x04 # 8-bit
+            self.data = [self.destination, self.source]
+        self.dataLength = len(self.data)
+        return super(PortAddress, self).encode()
+
+
+# Map of recognized IEIs
+IEI_CLASS_MAP = {0x00: Concatenation, # Concatenated short messages, 8-bit reference number
+                 0x08: Concatenation, # Concatenated short messages, 16-bit reference number
+                 0x04: PortAddress, # Application port addressing scheme, 8 bit address
+                 0x05: PortAddress # Application port addressing scheme, 16 bit address
+                }
+
+
+class Pdu(object):
+    """ Encoded SMS PDU. Contains raw PDU data and related meta-information """
+    
+    def __init__(self, data, tpduLength):
+        """ Constructor
+        @param data: the raw PDU data (as bytes)
+        @type data: bytearray
+        @param tpduLength: Length (in bytes) of the TPDU
+        @type tpduLength: int
+        """
+        self.data = data
+        self.tpduLength = tpduLength
+    
+    def __str__(self):
+        global PYTHON_VERSION
+        if PYTHON_VERSION < 3:
+            return str(self.data).encode('hex').upper()
+        else: #pragma: no cover
+            return str(codecs.encode(self.data, 'hex_codec'), 'ascii').upper() 
+
 
 def encodeSmsSubmitPdu(number, text, reference=0, validity=None, smsc=None, requestStatusReport=True, rejectDuplicates=False):
     """ Creates an SMS-SUBMIT PDU for sending a message with the specified text to the specified number
@@ -84,16 +248,11 @@ def encodeSmsSubmitPdu(number, text, reference=0, validity=None, smsc=None, requ
     @param rejectDuplicates: Flag that controls the TP-RD parameter (messages with same destination and reference may be rejected if True)
     @type rejectDuplicates: bool
             
-    @return: A tuple containing the SMS PDU as a bytearray, and the length of the TPDU part
-    @rtype: tuple
-    """ 
-    pdu = bytearray()
-    if smsc:
-        pdu.extend(_encodeAddressField(smsc, smscField=True))
-    else:
-        pdu.append(0x00) # Don't supply an SMSC number - use the one configured in the device
+    @return: A list of one or more tuples containing the SMS PDU (as a bytearray, and the length of the TPDU part
+    @rtype: list of tuples
+    """     
     tpduFirstOctet = 0x01 # SMS-SUBMIT PDU
-    if validity != None:    
+    if validity != None:
         # Validity period format (TP-VPF) is stored in bits 4,3 of the first TPDU octet
         if type(validity) == timedelta:
             # Relative (TP-VP is integer)
@@ -104,38 +263,90 @@ def encodeSmsSubmitPdu(number, text, reference=0, validity=None, smsc=None, requ
             tpduFirstOctet |= 0x18 # bit4 == 1, bit3 == 1
             validityPeriod = _encodeTimestamp(validity) 
         else:
-            raise ValueError('"validity" must be of type datetime.timedelta (for relative value) or datetime.datetime (for absolute value)')        
+            raise TypeError('"validity" must be of type datetime.timedelta (for relative value) or datetime.datetime (for absolute value)')        
     else:
         validityPeriod = None
     if rejectDuplicates:
         tpduFirstOctet |= 0x04 # bit2 == 1
     if requestStatusReport:
         tpduFirstOctet |= 0x20 # bit5 == 1
-    pdu.append(tpduFirstOctet)
-    pdu.append(reference) # message reference
-    # Add destination number    
-    pdu.extend(_encodeAddressField(number))
-    pdu.append(0x00) # Protocol identifier - no higher-level protocol
-    # Set data coding scheme based on text contents
+    
+    # Encode message text and set data coding scheme based on text contents
     try:
         encodedText = encodeGsm7(text)
     except ValueError:
         # Cannot encode text using GSM-7; use UCS2 instead
-        alphabet = 0x08 # UCS2        
-        encodedText = text.encode('utf-16')
-        userDataLength = len(encodedText) # Payload size in septets/characters
-        userData = encodedText
+        alphabet = 0x08 # UCS2
     else:
-        alphabet = 0x00 # GSM-7        
-        userDataLength = len(encodedText) # Payload size in septets/characters
-        userData = packSeptets(encodeGsm7(text))    
-    pdu.append(alphabet)
-    if validityPeriod:
-        pdu.extend(validityPeriod)
-    pdu.append(userDataLength)
-    pdu.extend(userData) # User Data (payload)
-    tpdu_length = len(pdu) - 1
-    return pdu, tpdu_length
+        alphabet = 0x00 # GSM-7    
+        
+    # Check if message should be concatenated
+    if len(text) > MAX_MESSAGE_LENGTH[alphabet]:
+        # Text too long for single PDU - add "concatenation" User Data Header
+        concatHeaderPrototype = Concatenation()
+        concatHeaderPrototype.reference = reference
+        pduCount = int(len(text) / MAX_MESSAGE_LENGTH[alphabet]) + 1
+        concatHeaderPrototype.parts  = pduCount
+        tpduFirstOctet |= 0x40
+    else:
+        concatHeaderPrototype = None
+        pduCount = 1
+    
+    # Construct required PDU(s)
+    pdus = []    
+    for i in xrange(pduCount):
+        pdu = bytearray()
+        if smsc:
+            pdu.extend(_encodeAddressField(smsc, smscField=True))
+        else:
+            pdu.append(0x00) # Don't supply an SMSC number - use the one configured in the device 
+    
+        udh = bytearray()
+        if concatHeaderPrototype != None:
+            concatHeader = copy(concatHeaderPrototype)
+            concatHeader.number = i + 1
+            pduText = text[i*153:(i+1) * 153]
+            udh.extend(concatHeader.encode())
+        else:
+            pduText = text
+        
+        udhLen = len(udh)        
+        
+        pdu.append(tpduFirstOctet)
+        pdu.append(reference) # message reference
+        # Add destination number    
+        pdu.extend(_encodeAddressField(number))
+        pdu.append(0x00) # Protocol identifier - no higher-level protocol
+    
+        pdu.append(alphabet)
+        if validityPeriod:
+            pdu.extend(validityPeriod)
+        
+        if alphabet == 0x00: # GSM-7
+            encodedText = encodeGsm7(pduText)
+            userDataLength = len(encodedText) # Payload size in septets/characters
+            if udhLen > 0:
+                shift = ((udhLen + 1) * 8) % 7 # "fill bits" needed to make the UDH end on a septet boundary
+                userData = packSeptets(encodedText, padBits=shift)
+                if shift > 0:
+                    userDataLength += 1 # take padding bits into account
+            else:
+                userData = packSeptets(encodedText)
+        elif alphabet == 0x08: # UCS2
+            userData = encodeUcs2(pduText)
+            userDataLength = len(userData)
+          
+        if udhLen > 0:            
+            userDataLength += udhLen + 1 # +1 for the UDH length indicator byte
+            pdu.append(userDataLength)
+            pdu.append(udhLen)
+            pdu.extend(udh) # UDH
+        else:
+            pdu.append(userDataLength)
+        pdu.extend(userData) # User Data (message payload)
+        tpdu_length = len(pdu) - 1
+        pdus.append(Pdu(pdu, tpdu_length))
+    return pdus
 
 def decodeSmsPdu(pdu):
     """ Decodes SMS pdu data and returns a tuple in format (number, text)
@@ -143,10 +354,16 @@ def decodeSmsPdu(pdu):
     @param pdu: PDU data as a hex string, or a bytearray containing PDU octects
     @type pdu: str or bytearray
     
+    @raise EncodingError: If the specified PDU data cannot be decoded
+    
     @return: The decoded SMS data as a dictionary
     @rtype: dict    
     """ 
-    pdu = toByteArray(pdu)
+    try:
+        pdu = toByteArray(pdu)
+    except Exception as e:
+        # Python 2 raises TypeError, Python 3 raises binascii.Error
+        raise EncodingError(e)
     result = {}
     pduIter = iter(pdu)
  
@@ -157,23 +374,16 @@ def decodeSmsPdu(pdu):
     tpduFirstOctet = next(pduIter) 
     
     pduType = tpduFirstOctet & 0x03 # bits 1-0
-    if pduType == 0x00: # SMS-DELIVER or SMS-DELIVER REPORT        
+    if pduType == 0x00: # SMS-DELIVER or SMS-DELIVER REPORT
         result['type'] = 'SMS-DELIVER'
         result['number'] = _decodeAddressField(pduIter)[0]
         result['protocol_id'] = next(pduIter)
         dataCoding = _decodeDataCoding(next(pduIter))
         result['time'] = _decodeTimestamp(pduIter)
         userDataLen = next(pduIter)
-        if dataCoding == 0x00: # GSM-7
-            userDataSeptets = unpackSeptets(pduIter, userDataLen)
-            result['text'] = decodeGsm7(userDataSeptets)
-        elif dataCoding == 0x02: # UCS2
-            userData = []
-            for i in xrange(userDataLen):
-                userData.append(chr(next(pduIter)))
-            result['text'] = ''.join(userData).decode('utf-16')
-        else:
-            result['text'] = ''        
+        udhPresent = (tpduFirstOctet & 0x40) != 0
+        ud = _decodeUserData(pduIter, userDataLen, dataCoding, udhPresent)
+        result.update(ud)
     elif pduType == 0x01: # SMS-SUBMIT or SMS-SUBMIT-REPORT
         result['type'] = 'SMS-SUBMIT'
         result['reference'] = next(pduIter) # message reference - we don't really use this
@@ -186,15 +396,9 @@ def decodeSmsPdu(pdu):
         elif validityPeriodFormat == 0x03: # TP-VP field present and semi-octet represented (absolute)            
             result['validity'] = _decodeTimestamp(pduIter)
         userDataLen = next(pduIter)
-        if dataCoding == 0x00: # GSM-7
-            userDataSeptets = unpackSeptets(pduIter, userDataLen)
-            result['text'] = decodeGsm7(userDataSeptets)
-        elif dataCoding == 0x02: # UCS2
-            for i in xrange(userDataLen):
-                userData.append(next(pduIter))
-            result['text'] = ''.join(userData).decode('utf-16')
-        else:
-            result['text'] = ''        
+        udhPresent = (tpduFirstOctet & 0x40) != 0
+        ud = _decodeUserData(pduIter, userDataLen, dataCoding, udhPresent)
+        result.update(ud)
     elif pduType == 0x02: # SMS-STATUS-REPORT or SMS-COMMAND
         result['type'] = 'SMS-STATUS-REPORT'
         result['reference'] = next(pduIter)
@@ -203,10 +407,46 @@ def decodeSmsPdu(pdu):
         result['discharge'] = _decodeTimestamp(pduIter)
         result['status'] = next(pduIter)        
     else:
-        raise EncodingError('Unknown SMS message type: {0}. First TPDU octect was: {1}'.format(pduType, tpduFirstOctet))
+        raise EncodingError('Unknown SMS message type: {0}. First TPDU octet was: {1}'.format(pduType, tpduFirstOctet))
     
     return result
 
+def _decodeUserData(byteIter, userDataLen, dataCoding, udhPresent):
+    """ Decodes PDU user data (UDHI (if present) and message text) """
+    result = {}
+    if udhPresent:
+        # User Data Header is present
+        result['udh'] = []
+        udhLen = next(byteIter)
+        ieLenRead = 0
+        # Parse and store UDH fields
+        while ieLenRead < udhLen:
+            ie = InformationElement.decode(byteIter)
+            ieLenRead += len(ie)
+            result['udh'].append(ie)
+        del ieLenRead
+        if dataCoding == 0x00: # GSM-7
+            # Since we are using 7-bit data, "fill bits" may have been added to make the UDH end on a septet boundary
+            shift = ((udhLen + 1) * 8) % 7 # "fill bits" needed to make the UDH end on a septet boundary
+            # Simulate another "shift" in the unpackSeptets algorithm in order to ignore the fill bits
+            prevOctet = next(byteIter)
+            shift += 1
+
+    if dataCoding == 0x00: # GSM-7
+        if udhPresent:
+            userDataSeptets = unpackSeptets(byteIter, userDataLen, prevOctet, shift)
+        else:
+            userDataSeptets = unpackSeptets(byteIter, userDataLen)
+        result['text'] = decodeGsm7(userDataSeptets)
+    elif dataCoding == 0x02: # UCS2
+        result['text'] = decodeUcs2(byteIter, userDataLen)
+    else: # 8-bit (data)
+        userData = []
+        for b in byteIter:
+            userData.append(unichr(b))
+        result['text'] = ''.join(userData)
+    return result
+
 def _decodeRelativeValidityPeriod(tpVp):
     """ Calculates the relative SMS validity period (based on the table in section 9.2.3.12 of GSM 03.40)
     @rtype: datetime.timedelta
@@ -219,7 +459,9 @@ def _decodeRelativeValidityPeriod(tpVp):
         return timedelta(days=(tpVp - 166))
     elif 197 <= tpVp <= 255:
         return timedelta(weeks=(tpVp - 192))
-    
+    else:
+        raise ValueError('tpVp must be in range [0, 255]')
+
 def _encodeRelativeValidityPeriod(validityPeriod):
     """ Encodes the specified relative validity period timedelta into an integer for use in an SMS PDU
     (based on the table in section 9.2.3.12 of GSM 03.40)
@@ -232,13 +474,15 @@ def _encodeRelativeValidityPeriod(validityPeriod):
     #seconds = validityPeriod.total_seconds()
     seconds = validityPeriod.seconds + (validityPeriod.days * 24 * 3600)
     if seconds <= 43200: # 12 hours
-        tpVp = (seconds / 300) - 1 # divide by 5 minutes, subtract 1
+        tpVp = int(seconds / 300) - 1 # divide by 5 minutes, subtract 1
     elif seconds <= 86400: # 24 hours
-        tpVp = (seconds - 43200) / 1800 + 143 # subtract 12 hours, divide by 30 minutes. add 143
+        tpVp = int((seconds - 43200) / 1800) + 143 # subtract 12 hours, divide by 30 minutes. add 143
     elif validityPeriod.days <= 30: # 30 days
         tpVp = validityPeriod.days + 166 # amount of days + 166
-    elif validityPeriod.days > 30:
-        tpVp = validityPeriod.days / 7 + 192 # amount of weeks + 192
+    elif validityPeriod.days <= 441: # max value of tpVp is 255
+        tpVp = int(validityPeriod.days / 7) + 192 # amount of weeks + 192
+    else:
+        raise ValueError('Validity period too long; tpVp limited to 1 octet (max value: 255)')
     return tpVp
         
 def _decodeTimestamp(byteIter):
@@ -258,10 +502,22 @@ def _encodeTimestamp(timestamp):
     @return: The encoded timestamp
     @rtype: bytearray
     """
-    if timestamp.utcoffset == None:
-        raise ValueError('Please specify a UTC offset for the timestamp (e.g. by using gsmmodem.util.SimpleOffsetTzInfo)')
-    dateStr = timestamp.strftime('%y%m%d%H%M%S%z')[-2]
-    return encodeSemiOctets(dateStr)    
+    if timestamp.tzinfo == None:
+        raise ValueError('Please specify time zone information for the timestamp (e.g. by using gsmmodem.util.SimpleOffsetTzInfo)')
+
+    # See if the timezone difference is positive/negative
+    tzDelta = timestamp.utcoffset()
+    if tzDelta.days >= 0:
+        tzValStr = '{0:0>2}'.format(int(tzDelta.seconds / 60 / 15))
+    else: # negative
+        tzVal = int((tzDelta.days * -3600 * 24 - tzDelta.seconds) / 60 / 15) # calculate offset in 0.25 hours
+        # Cast as literal hex value and set MSB of first semi-octet of timezone to 1 to indicate negative value
+        tzVal = int('{0:0>2}'.format(tzVal), 16) | 0x80
+        tzValStr = '{0:0>2X}'.format(tzVal)
+
+    dateStr = timestamp.strftime('%y%m%d%H%M%S') + tzValStr
+
+    return encodeSemiOctets(dateStr)
 
 def _decodeDataCoding(octet):
     if octet & 0xC0 == 0:
@@ -379,8 +635,8 @@ def decodeSemiOctets(encodedNumber, numberOfOctets=None):
     @rtype: string
     """
     number = []
-    if type(encodedNumber) == str:
-        encodedNumber = bytearray(encodedNumber.decode('hex'))
+    if type(encodedNumber) in (str, bytes):
+        encodedNumber = bytearray(codecs.decode(encodedNumber, 'hex_codec'))
     i = 0
     for octet in encodedNumber:        
         hexVal = hex(octet)[2:].zfill(2)   
@@ -436,7 +692,7 @@ def decodeGsm7(encodedText):
     """
     result = []
     if type(encodedText) == str:
-        encodedText = bytearray(encodedText)
+        encodedText = rawStrToByteArray(encodedText) #bytearray(encodedText)
     iterEncoded = iter(encodedText)
     for b in iterEncoded:
         if b == 0x1B: # ESC - switch to extended table
@@ -449,21 +705,24 @@ def decodeGsm7(encodedText):
             result.append(GSM7_BASIC[b])
     return ''.join(result)
 
-def packSeptets(octets):
+def packSeptets(octets, padBits=0):
     """ Packs the specified octets into septets
     
     Typically the output of encodeGsm7 would be used as input to this function. The resulting
     bytearray contains the original GSM-7 characters packed into septets ready for transmission.
     
     @rtype: bytearray
-    """    
+    """
     result = bytearray()    
     if type(octets) == str:
-        octets = iter(bytearray(octets))
+        octets = iter(rawStrToByteArray(octets))
     elif type(octets) == bytearray:
         octets = iter(octets)
-    shift = 0
-    prevSeptet = next(octets)
+    shift = padBits
+    if padBits == 0:
+        prevSeptet = next(octets)
+    else:
+        prevSeptet = 0x00
     for octet in octets:
         septet = octet & 0x7f;
         if shift == 7:
@@ -480,7 +739,7 @@ def packSeptets(octets):
         result.append(prevSeptet >> shift)
     return result
 
-def unpackSeptets(septets, numberOfSeptets=None):
+def unpackSeptets(septets, numberOfSeptets=None, prevOctet=None, shift=7):
     """ Unpacks the specified septets into octets 
     
     @param septets: Iterator or iterable containing the septets packed into octets
@@ -493,13 +752,11 @@ def unpackSeptets(septets, numberOfSeptets=None):
     """    
     result = bytearray()    
     if type(septets) == str:
-        septets = iter(bytearray(septets))
+        septets = iter(rawStrToByteArray(septets))
     elif type(septets) == bytearray:
         septets = iter(septets)    
     if numberOfSeptets == None:        
         numberOfSeptets = MAX_INT # Loop until StopIteration
-    shift = 7
-    prevOctet = None
     i = 0
     for octet in septets:
         i += 1
@@ -507,20 +764,19 @@ def unpackSeptets(septets, numberOfSeptets=None):
             shift = 1
             if prevOctet != None:                
                 result.append(prevOctet >> 1)            
-            #if len(result) < numberOfSeptets:
             if i <= numberOfSeptets:
                 result.append(octet & 0x7F)
                 prevOctet = octet                
-            #if len(result) == numberOfSeptets:
             if i == numberOfSeptets:
                 break
             else:
                 continue
         b = ((octet << shift) & 0x7F) | (prevOctet >> (8 - shift))
+        
         prevOctet = octet        
         result.append(b)
         shift += 1
-        #if len(result) == numberOfSeptets:
+        
         if i == numberOfSeptets:
             break
     if shift == 7:
@@ -529,3 +785,32 @@ def unpackSeptets(septets, numberOfSeptets=None):
             # The final septet value still needs to be unpacked
             result.append(b)        
     return result
+
+def decodeUcs2(byteIter, numBytes):
+    """ Decodes UCS2-encoded text from the specified byte iterator, up to a maximum of numBytes """
+    userData = []
+    i = 0
+    try:
+        while i < numBytes:
+            userData.append(unichr((next(byteIter) << 8) | next(byteIter)))
+            i += 2
+    except StopIteration:
+        # Not enough bytes in iterator to reach numBytes; return what we have
+        pass
+    return ''.join(userData)
+
+def encodeUcs2(text):
+    """ UCS2 text encoding algorithm
+    
+    Encodes the specified text string into UCS2-encoded bytes.
+    
+    @param text: the text string to encode
+    
+    @return: A bytearray containing the string encoded in UCS2 encoding
+    @rtype: bytearray
+    """
+    result = bytearray()
+    for b in map(ord, text):
+        result.append(b >> 8)
+        result.append(b & 0xFF)
+    return result
